import logging
from typing import Callable
from typing import Optional

from django.http import HttpRequest
from django.http import HttpResponse
from django.utils.module_loading import autodiscover_modules
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import View

from drakaina.rpc_protocols import BaseRPCProtocol
from drakaina.rpc_protocols import JsonRPCv2

log = logging.getLogger(__name__)


class RPCView(View):
    """Django class based view implements JSON-RPC"""

    http_method_names = ["post", "options"]
    handler: BaseRPCProtocol
    content_type: str

    @classmethod
    def as_view(
        cls,
        autodiscover: str = "rpc_methods",
        handler: Optional[BaseRPCProtocol] = None,
        **initkwargs,
    ) -> Callable:
        """

        :param autodiscover: submodule name(s) where defined RPC methods
        :type autodiscover: str
        :param handler:
        :type handler: BaseRPCProtocol
        :param initkwargs:
        :return: Instance of this class

        """
        cls.handler = handler or JsonRPCv2()
        cls.content_type = cls.handler.content_type
        view = super().as_view(**initkwargs)
        view.cls = cls

        # Scan specified sub-modules
        autodiscover_modules(autodiscover)

        return csrf_exempt(view)

    def http_method_not_allowed(
        self,
        request: HttpRequest,
        *args,
        **kwargs,
    ) -> HttpResponse:
        msg = f"HTTP Method Not Allowed ({request.method}): {request.path}"
        content = self.handler.get_raw_error(self.handler.BadRequestError(msg))

        return HttpResponse(content=content, content_type=self.content_type)

    def post(self, request: HttpRequest, *args, **kwargs) -> HttpResponse:
        if request.content_type != self.content_type or len(request.body) == 0:
            msg = f"HTTP Method Not Allowed ({request.method}): {request.path}"
            content = self.handler.get_raw_error(
                self.handler.BadRequestError(msg),
            )
        else:
            content = self.handler.handle_raw_request(
                request.body,
                request=request,
            )

        return HttpResponse(content=content, content_type=self.content_type)


# Warning! Do not commit this region
if __name__ == "__main__":
    django_view = RPCView.as_view()



from asyncio import iscoroutinefunction
from functools import update_wrapper
from typing import Any
from typing import Callable
from typing import Optional
from typing import TypeVar

from drakaina.registries import RPCRegistry

__all__ = (
    "RPCRegistry",
    "rpc_registry",
    "remote_procedure",
    "ENV_APP",
    "ENV_IS_AUTHENTICATED",
    "ENV_USER",
    "ENV_USER_ID",
    "ENV_JWT_PAYLOAD",
    "ENV_JWT_SCOPES",
)

# General registry
rpc_registry = RPCRegistry()

T = TypeVar("T")

# Environ (Request) context
ENV_APP = "drakaina.app"
ENV_IS_AUTHENTICATED = "drakaina.is_authenticated"
ENV_USER = "user"
ENV_USER_ID = "user_id"
ENV_JWT_PAYLOAD = "drakaina.jwt.payload"
ENV_JWT_SCOPES = "drakaina.jwt.scopes"


def remote_procedure(
    name: Optional[str] = None,
    registry: Optional[RPCRegistry] = None,
    provide_request: Optional[bool] = None,
    metadata: Optional[dict[str, Any]] = None,
    **meta_options,
) -> Callable:
    """Decorator allow wrap function and define it as remote procedure.

    :param name:
        Procedure name. Default as function name.
    :type name: str
    :param registry:
        Procedure registry custom object
    :type registry: RPCRegistry
    :param provide_request:
        Provide a request object or context data (from the transport layer).
        If `True`, then the request object or context can be supplied to
        the procedure as a `request` argument.
    :type provide_request: bool
    :param metadata:
        Metadata that can be processed by middleware.
    :type metadata: dict[str, Any]
    :param meta_options:
        Metadata that can be processed by middleware.

    """

    if callable(name):
        return __decorator(
            name,
            registry,
            None,
            provide_request,
            metadata=metadata,
            **meta_options,
        )
    elif not isinstance(name, (str, type(None))):
        raise TypeError(
            "Expected first argument to be an str, a callable, or None",
        )

    def decorator(procedure):
        assert callable(procedure)
        return __decorator(
            procedure,
            registry,
            name,
            provide_request,
            metadata=metadata,
            **meta_options,
        )

    return decorator


def __decorator(
    procedure: T,
    registry: RPCRegistry = None,
    name: str = None,
    provide_request: bool = None,
    **meta_options,
) -> T:
    """Returns a registered procedure"""

    if iscoroutinefunction(procedure):

        async def wrapper(*args, **kwargs):
            if not provide_request:
                if len(args) == 0:
                    kwargs.pop("request")
                else:
                    scope, *args = args
            return await procedure(*args, **kwargs)

    else:

        def wrapper(*args, **kwargs):
            if not provide_request:
                if len(args) == 0:
                    kwargs.pop("request")
                else:
                    environ, *args = args
            return procedure(*args, **kwargs)

    # Need to update the wrapper before registering in the registry
    decorated_procedure = update_wrapper(wrapper, procedure)

    _registry = registry if registry is not None else rpc_registry
    name = procedure.__name__ if name is None else name
    metadata = meta_options.pop("metadata") or {}
    _registry.register_procedure(
        decorated_procedure,
        name=name,
        provide_request=provide_request,
        metadata={**metadata, **meta_options},
    )

    return decorated_procedure



from drakaina.middlewares.base import BaseMiddleware
from drakaina.middlewares.cors import CORSMiddleware
from drakaina.middlewares.exception import ExceptionMiddleware
from drakaina.middlewares.request_wrapper import RequestWrapperMiddleware

__all__ = (
    "BaseMiddleware",
    "CORSMiddleware",
    "ExceptionMiddleware",
    "RequestWrapperMiddleware",
)



from typing import Any
from typing import Optional
from typing import Union

from drakaina._types import JSONRPCRequest
from drakaina._types import JSONRPCRequestObject
from drakaina._types import JSONRPCResponse
from drakaina._types import JSONRPCResponseObject
from drakaina.exceptions import AuthenticationFailedError
from drakaina.exceptions import BadRequestError
from drakaina.exceptions import DeserializationError
from drakaina.exceptions import ForbiddenError
from drakaina.exceptions import InternalServerError
from drakaina.exceptions import InvalidPermissionsError
from drakaina.exceptions import InvalidTokenError
from drakaina.exceptions import NotFoundError
from drakaina.exceptions import RPCError
from drakaina.exceptions import SerializationError
from drakaina.registries import RPCRegistry
from drakaina.rpc_protocols.base import BaseRPCProtocol
from drakaina.serializers import BaseSerializer
from drakaina.serializers import JsonSerializer


class JsonRPCError(RPCError):
    """JSON-RPC Common error

    Reserved for implementation-defined server-errors.
    Codes -32000 to -32099.

    """

    code: int = -32000
    default_message: str = "Server error"
    id: Union[int, str] = None
    data: Any = None

    def __init__(
        self,
        *args,
        message: str = None,
        id: Union[int, str] = None,
        data: Optional[Any] = None,
    ):
        super().__init__(*args)

        self.id = id
        if message:
            self.message = message

        if self.message and data:
            self.data = {"text": self.message.strip(), "details": data}
        elif self.message:
            self.data = self.message.strip()
        elif data:
            self.data = data

    def __str__(self) -> str:
        return f"{self.__class__.__name__} ({self.code} {self.default_message})"

    def as_dict(self) -> JSONRPCResponseObject:
        error = dict(
            jsonrpc="2.0",
            error={"code": self.code, "message": self.default_message},
            id=self.id,
        )

        if self.data:
            error["error"]["data"] = self.data

        return error


class InvalidRequestError(JsonRPCError):
    """Invalid Request

    The JSON sent is not a valid Request object.

    """

    code = -32600
    default_message = "Invalid Request"


class MethodNotFoundError(JsonRPCError):
    """Method not found

    The method does not exist / is not available.

    """

    code = -32601
    default_message = "Method not found"


class InvalidParamsError(JsonRPCError):
    """Invalid params

    Invalid method parameter(s).

    """

    code = -32602
    default_message = "Invalid params"


class InternalError(JsonRPCError):
    """Internal error

    Internal JSON-RPC error.

    """

    code = -32603
    default_message = "Internal error"


class ParseError(JsonRPCError):
    """Parse error

    Invalid JSON was received by the server.
    An error occurred on the server while parsing the JSON text.

    """

    code = -32700
    default_message = "Parse error"


# Implementation of Drakaina errors


class AuthenticationFailedJRPCError(JsonRPCError):
    """Authentication failed"""

    code = -32010
    default_message = "Authentication failed"


class InvalidTokenJRPCError(AuthenticationFailedJRPCError):
    """Invalid token error"""

    code = -32011
    default_message = "Invalid token error"


class ForbiddenJRPCError(AuthenticationFailedJRPCError):
    """Forbidden error"""

    code = -32012
    default_message = "Forbidden"


class InvalidPermissionsJRPCError(ForbiddenJRPCError):
    """Invalid permissions error"""

    code = -32013
    default_message = "Invalid permissions error"


class JsonRPCv2(BaseRPCProtocol):
    """JSON-RPC 2.0 implementation.

    :param registry:
        Registry of remote procedures.
        Default: `drakaina.registries.rpc_registry` (generic module instance)
    :type registry: RPCRegistry
    :param serializer:
        Serializer object. Default: `JsonSerializer` (stdlib.json)
    :type serializer: BaseSerializer

    """

    def __init__(
        self,
        registry: Optional[RPCRegistry] = None,
        serializer: Optional[BaseSerializer] = None,
    ):
        if serializer is None:
            serializer = JsonSerializer()
        super().__init__(registry=registry, serializer=serializer)

        self._errors_map = {
            Exception: JsonRPCError,
            RPCError: JsonRPCError,
            BadRequestError: InvalidRequestError,
            NotFoundError: MethodNotFoundError,
            # None: InvalidParamsError,
            InternalServerError: InternalError,
            SerializationError: InternalError,
            DeserializationError: ParseError,
            AuthenticationFailedError: AuthenticationFailedJRPCError,
            InvalidTokenError: InvalidTokenJRPCError,
            ForbiddenError: ForbiddenJRPCError,
            InvalidPermissionsError: InvalidPermissionsJRPCError,
        }

    def handle(
        self,
        rpc_request: JSONRPCRequest,
        request: Optional[Any] = None,
    ) -> Optional[JSONRPCResponse]:
        """Handles a procedure call or batch of procedure call

        :param rpc_request:
            RPC request in protocol format.
        :type rpc_request: JSONRPCRequest
        :param request:
            Optional parameter that can be passed as an
            argument to the procedure.
        :type request: Any
        :return:
            Returns the result in protocol format.
        :rtype: JSONRPCResponse

        """
        # Check bad request
        if not (isinstance(rpc_request, (dict, list)) and len(rpc_request) > 0):
            return InvalidRequestError().as_dict()

        # Handle batch request
        if isinstance(rpc_request, list):
            batch_result = []
            for request_object in rpc_request:
                result = self.execute(request_object, request=request)
                if result is not None:
                    batch_result.append(result)
            if len(batch_result) == 0:
                return None
            return batch_result

        # Handle single request
        return self.execute(rpc_request, request=request)

    def execute(
        self,
        procedure_call: JSONRPCRequestObject,
        request: Optional[Any] = None,
    ) -> Optional[JSONRPCResponseObject]:
        """Execute a remote procedure call.

        :param procedure_call:
            RPC request object in protocol format.
        :type procedure_call: JSONRPCRequestObject
        :param request:
            Optional parameter that can be passed as an
            argument to the procedure. By default, None will be passed.
        :type request: Any
        :return:
            Returns a result object in protocol format.
        :rtype: JSONRPCResponseObject

        """
        if not isinstance(procedure_call, dict):
            return InvalidRequestError().as_dict()
        method: str = procedure_call.get("method")
        params: Optional[Union[list, dict]] = procedure_call.get("params")
        request_id: Optional[Union[int, str]] = procedure_call.get("id")

        # Validate protocol
        if (
            procedure_call.get("jsonrpc") != "2.0"
            or not isinstance(method, str)
            or not (params is None or isinstance(params, (list, dict)))
            or not (request_id is None or isinstance(request_id, (int, str)))
        ):
            return InvalidRequestError(id=request_id).as_dict()

        # Getting procedure
        procedure = self.registry[method]
        if procedure is None:
            if request_id is None:
                return None
            return MethodNotFoundError(id=request_id).as_dict()

        # Prepare parameters
        args, kwargs = (), {}
        if params is not None:
            if isinstance(params, list):
                args = (request, *params)
            elif isinstance(params, dict):
                kwargs = dict(request=request, **params)
        else:
            kwargs = dict(request=request)

        # Execute RPC method
        try:
            # todo: what if method is coroutine? use asgi_ref.async_to_sync
            result = procedure(*args, **kwargs)
        except JsonRPCError as err:
            err.id = request_id
            return err.as_dict()
        except Exception as err:
            return InternalError(message=str(err), id=request_id).as_dict()

        if request_id is None:
            return None

        return dict(jsonrpc="2.0", result=result, id=request_id)

    @property
    def default_error(self):
        return InternalError

    def smd_scheme(self):
        return {}



"""Submodule with implementations support of JSON Web Tokens

Standard: https://www.rfc-editor.org/rfc/rfc7519
"""
from __future__ import annotations

from asyncio import iscoroutinefunction
from functools import update_wrapper
from typing import Callable
from typing import Iterable
from typing import Optional
from typing import TypeVar

from drakaina import ENV_IS_AUTHENTICATED
from drakaina import ENV_JWT_SCOPES
from drakaina._types import ASGIScope
from drakaina._types import WSGIEnvironment
from drakaina.contrib.jwt.errors import InvalidJWTTokenError
from drakaina.contrib.jwt.types import Comparator
from drakaina.exceptions import ForbiddenError
from drakaina.registries import is_rpc_procedure
from drakaina.utils import iterable_str_arg

__all__ = (
    "check_permissions",
    "login_required",
    "match_all",
    "match_any",
    "RESERVED_FIELDS",
    "SUPPORTED_ALGORITHMS",
)

RESERVED_FIELDS = ("iss", "sub", "aud", "exp", "nbf", "iat", "jti")
SUPPORTED_ALGORITHMS = ("HS256", "HS384", "HS512", "RS256", "RS384", "RS512")

T = TypeVar("T")


def login_required(*_args) -> Callable:
    """Requires login decorator.

    Gives access to the procedure only to authenticated users.

    """

    def __decorator(procedure: T) -> T:
        """Returns a registered procedure"""
        if not is_rpc_procedure(procedure):
            raise TypeError(
                "Incorrect usage of decorator. Please use "
                "the `drakaina.remote_procedure` decorator first.",
            )

        if iscoroutinefunction(procedure):

            async def wrapper(*args, **kwargs):
                if len(args) == 0:
                    scope: ASGIScope = kwargs.get("request")
                else:
                    scope: ASGIScope = args[0]

                if not scope.get(ENV_IS_AUTHENTICATED, False):
                    raise ForbiddenError("Authorization required")
                return await procedure(*args, **kwargs)

        else:

            def wrapper(*args, **kwargs):
                if len(args) == 0:
                    environ: WSGIEnvironment = kwargs.get("request")
                else:
                    environ: WSGIEnvironment = args[0]

                if not environ.get(ENV_IS_AUTHENTICATED, False):
                    raise ForbiddenError("Authorization required")
                return procedure(*args, **kwargs)

        return update_wrapper(wrapper, procedure)

    if len(_args) > 0:
        if callable(_args[0]):
            return __decorator(_args[0])
        else:
            raise TypeError("Expected first argument to be a callable")

    def decorator(procedure):
        assert callable(procedure)
        return __decorator(procedure)

    return decorator


def match_any(
    required: Iterable[str],
    provided: str | Iterable[str],
) -> bool:
    return any((scope in provided for scope in required))


def match_all(
    required: Iterable[str],
    provided: str | Iterable[str],
) -> bool:
    return set(required).issubset(set(provided))


def check_permissions(
    scopes: str | Iterable[str],
    comparator: Comparator = match_all,
) -> Callable:
    """Permission decorator.

    Gives access to the procedure only to authorized users.

    """
    if not callable(comparator):
        raise TypeError("comparator should be a func")

    procedure_scopes = iterable_str_arg(scopes)

    def decorator(procedure: T) -> T:
        if not is_rpc_procedure(procedure):
            raise TypeError(
                "Incorrect usage of decorator. Please use "
                "the `drakaina.remote_procedure` decorator first.",
            )

        if iscoroutinefunction(procedure):

            async def wrapper(*args, **kwargs):
                if len(args) == 0:  # noqa
                    scope: ASGIScope = kwargs.get("request")
                else:
                    scope: ASGIScope = args[0]

                if not scope.get(ENV_IS_AUTHENTICATED, False):
                    raise ForbiddenError("Authorization required")

                user_scopes = _get_scopes(scope)
                if not isinstance(user_scopes, Iterable):
                    raise InvalidJWTTokenError("Invalid permissions format")

                if not comparator(procedure_scopes, user_scopes):
                    raise ForbiddenError("Forbidden")

                return await procedure(*args, **kwargs)

        else:

            def wrapper(*args, **kwargs):
                if len(args) == 0:  # noqa
                    environ: WSGIEnvironment = kwargs.get("request")
                else:
                    environ: WSGIEnvironment = args[0]

                if not environ.get(ENV_IS_AUTHENTICATED, False):
                    raise ForbiddenError("Authorization required")

                user_scopes = _get_scopes(environ)
                if not isinstance(user_scopes, Iterable):
                    raise InvalidJWTTokenError("Invalid permissions format")

                if not comparator(procedure_scopes, user_scopes):
                    raise ForbiddenError("Forbidden")

                return procedure(*args, **kwargs)

        return update_wrapper(wrapper, procedure)

    return decorator


def _get_scopes(
    request: WSGIEnvironment | ASGIScope,
) -> Optional[Iterable[str]]:
    scopes = request.get(ENV_JWT_SCOPES)
    if scopes:
        return iterable_str_arg(scopes)



"""
https://github.com/rizki4106/django_jwt_middleware/blob/master/src/django_jwt_middleware/source.py
"""

from django.http import JsonResponse
import jwt


class JwtMiddleWare:
    def with_cookie(self, cookie_name="token", jwt_key="", algorithm="HS256"):
        """
        Verify json web token from cookie this decorator takes 3 arguments :
        1. cookie_name -> string -> cookie name where you store the token
        2. jwt_key -> string -> jwt secret key
        3. algorithm -> string -> default is HS256
        """

        def decorator(views_func):
            def container(request, *args, **kwargs):
                cookies = request.COOKIES
                # if cookie exists will continue to check token
                # if not exists will return 403 http status code
                if cookie_name in cookies:

                    # if token valid will continue to the next request
                    # if not valid will send 403 http status code
                    try:
                        jwt.decode(
                            str(cookies.get(cookie_name)), jwt_key, algorithm
                        )
                        return views_func(request, *args, **kwargs)

                    except Exception as err:

                        response = JsonResponse(
                            data={"status": 403, "message": str(err)}
                        )
                        response.status_code = 403
                        return response
                else:

                    response = JsonResponse(
                        data={
                            "status": 403,
                            "message": "token cookie must be sent in request cookie as  {}".format(
                                cookie_name
                            ),
                        }
                    )
                    response.status_code = 403
                    return response

            return container

        return decorator

    def with_header(self, header_name="token", jwt_key="", algorithm="HS256"):
        """
        Verify json web token from cookie this decorator takes 3 arguments :
        1. header_name -> string -> header name where you store the token
        2. jwt_key -> string -> jwt secret key
        3. algorithm -> string -> default is HS256
        """

        def decorator(views_func):
            def container(request, *args, **kwargs):
                headers = request.headers
                # if header exists will continue to check token
                # if not exists will return 403 http status code
                if header_name.lower() in headers:

                    # if token valid will continue to the next request
                    # if not valid will send 403 http status code
                    try:
                        jwt.decode(
                            str(headers.get(header_name.lower())),
                            jwt_key,
                            algorithm,
                        )
                        return views_func(request, *args, **kwargs)

                    except Exception as err:

                        response = JsonResponse(
                            data={"status": 403, "message": str(err)}
                        )
                        response.status_code = 403
                        return response
                else:

                    response = JsonResponse(
                        data={
                            "status": 403,
                            "message": "token headers must be sent in request header as {}".format(
                                header_name.lower()
                            ),
                        }
                    )
                    response.status_code = 403
                    return response

            return container

        return decorator
